/* $XFree86: xc/programs/Xserver/hw/xfree86/drivers/mga/mga_storm.c,v 1.64 2000/02/12 20:45:24 dawes Exp $ */


/* All drivers should typically include these */
#include "xf86.h"
#include "xf86_OSproc.h"
#include "xf86_ansic.h"

/* For correct __inline__ usage */
#include "compiler.h"

/* Drivers that need to access the PCI config space directly need this */
#include "xf86Pci.h"

/* Drivers for PCI hardware need this */
#include "xf86PciInfo.h"

/* Drivers that use XAA need this */
#include "xaa.h"
#include "xaalocal.h"
#include "xf86fbman.h"
#include "miline.h"
#include "servermd.h"

#include "mq.h"
#include "mq2hw.h"
#include "mq_ge.h"


/*
#ifdef XF86DRI
void MQDRIInitBuffers(WindowPtr pWin, 
			     RegionPtr prgn, CARD32 index);
void MQDRIMoveBuffers(WindowPtr pParent, DDXPointRec ptOldOrg,
			     RegionPtr prgnSrc, CARD32 index);
#endif
*/

#ifdef CHECK_24BPP
static Bool
MQAccelInit24(ScrnInfoPtr pScrn, MQPtr pMQ) 
{
    XAAInfoRecPtr infoPtr;
    infoPtr = pMQ->AccelInfoRec;

    infoPtr->Flags = MICROSOFT_ZERO_LINE_BIAS
		| LINEAR_FRAMEBUFFER
		| NO_PLANEMASK
		| NO_TRANSPARENCY
                | OFFSCREEN_PIXMAPS
		| PIXMAP_CACHE;
 
    infoPtr->Sync = MQSync;
    infoPtr->SetupForScreenToScreenCopy = MQ24SetupForScreenToScreenCopy;
    infoPtr->SubsequentScreenToScreenCopy =
        	MQ24SubsequentScreenToScreenCopy;
    infoPtr->SetupForSolidFill = MQ24SetupForSolidFill;
    infoPtr->SubsequentSolidFillRect = MQ24SubsequentSolidFillRect;

    infoPtr->Sync = MQSync;

    /* Screen To Screen */
    infoPtr->ScreenToScreenCopyFlags = NO_TRANSPARENCY | NO_PLANEMASK;
    infoPtr->SetupForScreenToScreenCopy = MQ24SetupForScreenToScreenCopy;
    infoPtr->SubsequentScreenToScreenCopy = MQ24SubsequentScreenToScreenCopy;

    /* solid Fill - Rectangle */
    infoPtr->SolidFillFlags = NO_TRANSPARENCY | NO_PLANEMASK;
    infoPtr->SetupForSolidFill = MQ24SetupForSolidFill;
    infoPtr->SubsequentSolidFillRect = MQ24SubsequentSolidFillRect;
    infoPtr->SubsequentSolidFillTrap = NULL;

    /* CPU to Screen - color image transfer */
    infoPtr->ImageWriteFlags =
                        CPU_TRANSFER_PAD_QWORD |
			| SCANLINE_PAD_DWORD    //HSU: default & the only option
			| LEFT_EDGE_CLIPPING 
			| SYNC_AFTER_IMAGE_WRITE
                        | ROP_NEEDS_SOURCE
			| NO_GXCOPY		//HSU: skip ge if small image
                        | NO_TRANSPARENCY
                        | NO_PLANEMASK
                        ;
    infoPtr->ImageWriteRange = 0x4000; //we can combine PSF and SSF together
    infoPtr->ImageWriteBase = pMQ->IOBase + PSF_BASE;
    infoPtr->SetupForImageWrite = MQ24SetupForImageWrite;
    infoPtr->SubsequentImageWriteRect = MQ24SubsequentImageWrite;

    /* clipping */
    infoPtr->SetClippingRectangle = MQ24SetClippingRectangle;
    infoPtr->DisableClipping = MQ24DisableClipping;
    infoPtr->ClippingFlags =  HARDWARE_CLIP_SOLID_LINE
                            | HARDWARE_CLIP_SCREEN_TO_SCREEN_COPY 
                            | HARDWARE_CLIP_SCREEN_TO_SCREEN_COLOR_EXPAND 
                          //| HARDWARE_CLIP_DASHED_LINE
                            |  HARDWARE_CLIP_SOLID_FILL 
                            | HARDWARE_CLIP_MONO_8x8_FILL
      //HSU: Can't put in color 8x8 pat yet | HARDWARE_CLIP_COLOR_8x8_FILL
                          ;
}
#endif	//CHECK_24BPP

Bool
MQAccelInitAll(ScrnInfoPtr pScrn, MQPtr pMQ)
{
    XAAInfoRecPtr infoPtr;

    infoPtr = pMQ->AccelInfoRec;
    infoPtr->Flags = MICROSOFT_ZERO_LINE_BIAS
		| LINEAR_FRAMEBUFFER
		//| NO_PLANEMASK
		//| NO_TRANSPARENCY
                | OFFSCREEN_PIXMAPS
		| PIXMAP_CACHE;
 
    //infoPtr->CachePixelGranularity = 128;
    infoPtr->Sync = MQSync;

    /* Screen To Screen */
    infoPtr->ScreenToScreenCopyFlags = NO_TRANSPARENCY | NO_PLANEMASK;
    infoPtr->SetupForScreenToScreenCopy = MQSetupForScreenToScreenCopy;
    infoPtr->SubsequentScreenToScreenCopy = MQSubsequentScreenToScreenCopy;

    /* solid Fill - Rectangle */
    infoPtr->SolidFillFlags = NO_TRANSPARENCY; 
    infoPtr->SetupForSolidFill = MQSetupForSolidFill;
    infoPtr->SubsequentSolidFillRect = MQSubsequentSolidFillRect;
    infoPtr->SubsequentSolidHorVertLine = MQSubsequentSolidHorVertLine;
    infoPtr->SubsequentSolidFillTrap = NULL;

    /* solid Fill - line */
    infoPtr->SolidLineFlags = NO_PLANEMASK | NO_TRANSPARENCY;
    infoPtr->SetupForSolidLine = MQSetupForSolidFill;
    infoPtr->SubsequentSolidBresenhamLine = MQSubsequentSolidBresenhamLine;
/*
    infoPtr->PolySegmentThinSolid = MQVPolySegmentThinSolidWrapper;
    infoPtr->PolylinesThinSolid = MQVPolylinesThinSolidWrapper;
*/

    //Default is already 16 bits for bresenham error term bits, and
    //MQ200 is capable of doing that.
    //infoPtr->SolidBresenhamLineErrorTermBits = 16;

    /* mono 8x8 pattern */
    //HSU: We don't use scrn origin for our pattern offset.
    //     So do not set HARDWARE_PATTERN_SCREEN_ORIGIN flag
    //HSU: We don't support 8x8 color pattern sitting in offscreen.
    //     We can read from offscreen back to our color pattern registers
    //     but too costly.  Besides XAA claims that if mono 8x8 pattern
    //     is supported, we can skip 8x8 color pattern fill as they are
    //     not that common.
    infoPtr->Mono8x8PatternFillFlags = HARDWARE_PATTERN_PROGRAMMED_BITS
					| HARDWARE_PATTERN_PROGRAMMED_ORIGIN
					| BIT_ORDER_IN_BYTE_MSBFIRST
                                        //| NO_PLANEMASK
                                        ; 
    infoPtr->SetupForMono8x8PatternFill = MQSetupForMono8x8PatternFill;
    infoPtr->SubsequentMono8x8PatternFillRect =
                                        MQSubsequentMono8x8PatternFillRect;

    /* CPU to Screen - 1bpp expansion */
    infoPtr->CPUToScreenColorExpandFillFlags =
                          CPU_TRANSFER_PAD_QWORD //HSU:Let XAA pad till 64bit
			| SCANLINE_PAD_DWORD     //HSU:default & the only option
			| BIT_ORDER_IN_BYTE_MSBFIRST
			| LEFT_EDGE_CLIPPING
			//| LEFT_EDGE_CLIPPING_NEGATIVE_X  //HSU: no support
			| SYNC_AFTER_COLOR_EXPAND
                        | ROP_NEEDS_SOURCE        //HSU: check s3virge
                        //| NO_TRANSPARENCY
                        //| NO_PLANEMASK
                        ;
    infoPtr->ColorExpandRange = 0x4000; //we can combine PSF and SSF together
    infoPtr->ColorExpandBase = pMQ->IOBase + PSF_BASE;
    infoPtr->SetupForCPUToScreenColorExpandFill =
		MQSetupForCPUToScreenColorExpandFill;
    infoPtr->SubsequentCPUToScreenColorExpandFill =
		MQSubsequentCPUToScreenColorExpandFill;

    /* CPU to Screen - color image transfer */
    infoPtr->ImageWriteFlags =
                        CPU_TRANSFER_PAD_QWORD |
			SCANLINE_PAD_DWORD |    //HSU: default & the only option
			LEFT_EDGE_CLIPPING |
			SYNC_AFTER_IMAGE_WRITE |
                        ROP_NEEDS_SOURCE |
			NO_GXCOPY |  //HSU: skip ge if small image
                        NO_TRANSPARENCY //|
                        //NO_PLANEMASK
                        ;
    infoPtr->ImageWriteRange = 0x4000; //we can combine PSF and SSF together
    infoPtr->ImageWriteBase = pMQ->IOBase + PSF_BASE;
    infoPtr->SetupForImageWrite = MQSetupForImageWrite;
    infoPtr->SubsequentImageWriteRect = MQSubsequentImageWrite;

    /* screen to screen color expansion */
    infoPtr->ScreenToScreenColorExpandFillFlags =
    			BIT_ORDER_IN_BYTE_MSBFIRST
    			| NO_PLANEMASK
			| ROP_NEEDS_SOURCE
			;
    infoPtr->SetupForScreenToScreenColorExpandFill = 
		        MQSetupForScreenToScreenColorExpandFill;
    infoPtr->SubsequentScreenToScreenColorExpandFill = 
	                MQSubsequentScreenToScreenColorExpandFill;
 
    /* clipping */
    //HSU: It looks like Linux is looking for real destination
    //     FrameBuffer clipping that works for all GE and FB writes.
    //     Can't set all clipping flags here for MQ200 GE that we
    //     support unless we intercept and use GE for all rectangle
    //     fill related operations etc.
    infoPtr->SetClippingRectangle = MQSetClippingRectangle;
    infoPtr->DisableClipping = MQDisableClipping;
    infoPtr->ClippingFlags = HARDWARE_CLIP_SCREEN_TO_SCREEN_COPY 
                          //| HARDWARE_CLIP_SOLID_LINE
                          //| HARDWARE_CLIP_SCREEN_TO_SCREEN_COLOR_EXPAND 
                          //| HARDWARE_CLIP_DASHED_LINE
                          //|  HARDWARE_CLIP_SOLID_FILL 
                            | HARDWARE_CLIP_MONO_8x8_FILL
      			  //| HARDWARE_CLIP_COLOR_8x8_FILL
                          ;

#if 0   //The followings are for future improvement
    infoPtr->FillCacheBltRects = MQFillCacheBltRects;

    /* dashed lines */
    infoPtr->DashedLineFlags = LINE_PATTERN_MSBFIRST_LSBJUSTIFIED;
    infoPtr->SetupForDashedLine = MQSetupForDashedLine;
    infoPtr->SubsequentDashedTwoPointLine =
		MQSubsequentDashedTwoPointLine;
    infoPtr->DashPatternMaxLength = 128;

    /* midrange replacements */
    if(infoPtr->SetupForCPUToScreenColorExpandFill && 
    			infoPtr->SubsequentCPUToScreenColorExpandFill)
    {
        infoPtr->FillColorExpandRects = MQFillColorExpandRects; 

    	infoPtr->WriteBitmapFlags |= NO_PLANEMASK;
        infoPtr->WriteBitmap = MQWriteBitmapColorExpand;

        infoPtr->TEGlyphRenderer = MQTEGlyphRenderer;  
        if(BITMAP_SCANLINE_PAD == 32)
	    infoPtr->NonTEGlyphRenderer = MQNonTEGlyphRenderer;  
    }

    //HSU: Nice if we support PCI master.  Not in MQ200 yet!!!
    //if(pMga->ILOADBase && pMga->UsePCIRetry && infoPtr->SetupForSolidFill)
    //{
    //	infoPtr->FillSolidRects = MQFillSolidRectsDMA;
    //	infoPtr->FillSolidSpans = MQFillSolidSpansDMA;
    //}

    if(infoPtr->SetupForSolidFill)
    {
	infoPtr->ValidatePolyArc = MQValidatePolyArc;
	infoPtr->PolyArcMask = GCFunction | GCLineWidth | GCPlaneMask | 
				GCLineStyle | GCFillStyle;
	infoPtr->ValidatePolyPoint = MQValidatePolyPoint;
	infoPtr->PolyPointMask = GCFunction | GCPlaneMask;
    }

    //Corresponding flags for each improved functions
    infoPtr->FillSolidRectsFlags |= NO_PLANEMASK;
    infoPtr->DashedLineFlags |= NO_PLANEMASK;
    infoPtr->FillColorExpandRectsFlags |= NO_PLANEMASK; 
    infoPtr->FillSolidSpansFlags |= NO_PLANEMASK;
    infoPtr->FillMono8x8PatternRectsFlags |= NO_PLANEMASK;
    infoPtr->NonTEGlyphRendererFlags |= NO_PLANEMASK;
    infoPtr->TEGlyphRendererFlags |= NO_PLANEMASK;
    infoPtr->FillCacheBltRectsFlags |= NO_PLANEMASK;
#endif  //Improvements

    return TRUE;
}

static Bool
MQCreateInfoRec( MQPtr pMQ )
{
    pMQ->AccelInfoRec = XAACreateInfoRec();
    if ( pMQ->AccelInfoRec == NULL )
	ErrorF("XAACreateInfoRec: Fail!\n");
    return (pMQ->AccelInfoRec != NULL);
}

Bool 
MQAccelInit(ScreenPtr pScreen)
{ 
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    MQPtr 	pMQ = MQPTR(pScrn);
    BoxRec AvailFBArea;
#ifdef XF86DRI
    RegionRec MemRegion;
#endif
    BOOL	status;

    ErrorF("MQAccelInit: %d\n", pScrn->bitsPerPixel);

    switch( pScrn->bitsPerPixel )
    {
        case 8:
	    MQCreateInfoRec( pMQ );
            geWAITCMDFIFO( 1 );
	    geREG(DEST_STRIDE, (pScrn->virtualX | GE_8BPP));
	    pMQ->BppShift = 0;
	    pMQ->BppBitMask = 0x07;
	    pMQ->FullPlaneMask = 0xFF;
    	    status = MQAccelInitAll(pScrn, pMQ);
	    break;
        case 16:
	    MQCreateInfoRec( pMQ );
            geWAITCMDFIFO( 1 );
	    geREG(DEST_STRIDE, ((pScrn->virtualX * 2) | GE_16BPP));
	    pMQ->BppShift = 1;
	    pMQ->BppBitMask = 0x0F;
	    pMQ->FullPlaneMask = 0xFFFF;
    	    status = MQAccelInitAll(pScrn, pMQ);
	    break;
    #ifdef CHECK_24BPP
        case 24:
	    MQCreateInfoRec( pMQ );
	    //Treat 24bpp GE like a 8bpp one.
            geWAITCMDFIFO( 1 );
	    geREG(DEST_STRIDE, ((pScrn->virtualX * 3) | GE_8BPP));
	    pMQ->BppShift = 0;
	    pMQ->BppBitMask = 0x07;
	    pMQ->FullPlaneMask = 0xFF;
    	    status = MQAccelInit24(pScrn, pMQ);
	    break;
    #endif //CHECK_24BPP
        case 32:
	    MQCreateInfoRec( pMQ );
            geWAITCMDFIFO( 1 );
	    geREG(DEST_STRIDE, ((pScrn->virtualX * 4) | GE_32BPP));
	    pMQ->BppShift = 2;
	    pMQ->BppBitMask = 0x1F;
	    pMQ->FullPlaneMask = 0xFFFFFFFF;
    	    status = MQAccelInitAll(pScrn, pMQ);
	    break;
	default:
	    return FALSE;	//Not supported color depth. Return immed.
    }

    if ( !status )
    {
        ErrorF("MQAccelInit Subs Failed!!!\n");
	return status;
    }

    //Initialize GE registers and global variable in MQ structure.
    geWAITCMDFIFO( 1 );
    geREG( BASE_ADDRESS, 0);		//set GE base address to 0
    pMQ->SrcImageDataAddr = (ULONG *)(pMQ->IOBase + SRC_IMAGE_DATA);
    pMQ->Need64BitDummy = 0;
    pMQ->AccelFlags = 0;

    //pMQ->CPUControl = cpuREAD(CPU_CONTROL);

    ErrorF( "BASE_ADDRESS=0x%.8x\n", geREAD(BASE_ADDRESS) );
    ErrorF( "DEST_STRIDE=0x%.8x\n", geREAD(DEST_STRIDE) );

    AvailFBArea.x1 = 0;
    AvailFBArea.x2 = pScrn->displayWidth;

#ifndef XF86DRI //By default XF86DRI flag is not defined
    AvailFBArea.y1 = 0;
    AvailFBArea.y2 = (min(pMQ->FbUsableSize, MQ200_FB_SIZE)) / 
		(pScrn->displayWidth * pScrn->bitsPerPixel / 8);
    status = xf86InitFBManager(pScreen, &AvailFBArea); 
    ErrorF("xf86InitFBManager: %d, %d\n", AvailFBArea.y2, status);
#else
    AvailFBArea.y1 = pScrn->virtualY;
    //HSU: Why + 128?? AvailFBArea.y2 = pScrn->virtualY+128;
    AvailFBArea.y2 = pScrn->virtualY;   //Don't include any offscreen yet

    REGION_INIT(pScreen, &MemRegion, &AvailFBArea, 1);
    status = xf86InitFBManagerRegion(pScreen, &MemRegion);
    ErrorF("xf86InitFBManager: %d, %d\n", AvailFBArea.y2, status);
    REGION_UNINIT(pScreen, &MemRegion);
#endif

    status = XAAInit(pScreen, pMQ->AccelInfoRec);
    ErrorF( "XAAInit: %d\n", status );

    return status;
}
