/* $XFree86: $ */

#include "xf86.h"
#include "xf86_OSproc.h"
#include "xf86_ansic.h"
#include "xf86Pci.h"
#include "xf86PciInfo.h"
#include "xaa.h"
#include "xaalocal.h"
#include "mq.h"
#include "mq2hw.h"
#include "mqproto.h"
#include "dgaproc.h"


static Bool MQ_OpenFramebuffer(ScrnInfoPtr, char **, unsigned char **, 
					int *, int *, int *);
static Bool MQ_SetMode(ScrnInfoPtr, DGAModePtr);
static int  MQ_GetViewport(ScrnInfoPtr);
static void MQ_SetViewport(ScrnInfoPtr, int, int, int);
static void MQ_FillRect(ScrnInfoPtr, int, int, int, int, unsigned long);
static void MQ_BlitRect(ScrnInfoPtr, int, int, int, int, int, int);
static void MQ_BlitTransRect(ScrnInfoPtr, int, int, int, int, int, int, 
					unsigned long);

static
DGAFunctionRec MQ_DGAFuncs =
{
   MQ_OpenFramebuffer,
   NULL,
   MQ_SetMode,
   MQ_SetViewport,
   MQ_GetViewport,
   MQSync,
   MQ_FillRect,
   MQ_BlitRect,
   NULL        //HSU: no support for transparent screen-to-screen
               //MQ_BlitTransRect
};


static int
FindSmallestPitch(
   MQPtr pMQ,
   int width
)
{
   int Pitches[] = 
	{640, 800, 1024, 1152, 1280, 0};
   int *linePitches;

   linePitches = Pitches;
   while( *linePitches < width )
	linePitches++;
   return *linePitches;
}

static DGAModePtr
MQSetupDGAMode(
   ScrnInfoPtr pScrn,
   DGAModePtr modes,
   int *num,
   int bitsPerPixel,
   int depth,
   Bool pixmap,
   int secondPitch,
   unsigned long red,
   unsigned long green,
   unsigned long blue,
   short visualClass
)
{
   DisplayModePtr firstMode, pMode;
   MQPtr pMQ = MQPTR(pScrn);
   DGAModePtr mode, newmodes;
   int size, pitch, Bpp = bitsPerPixel >> 3;

SECOND_PASS:

   pMode = firstMode = pScrn->modes;

   while(1)
   {
	pitch = FindSmallestPitch(pMQ, pMode->HDisplay);
	size = pitch * Bpp * pMode->VDisplay;

	if((!secondPitch || (pitch != secondPitch)) &&
		(size <= pMQ->FbUsableSize))
        {
	    if(secondPitch)
		pitch = secondPitch; 

	    if(!(newmodes = xrealloc(modes, (*num + 1) * sizeof(DGAModeRec))))
		break;

	    modes = newmodes;
	    mode = modes + *num;

	    mode->mode = pMode;
	    mode->flags = DGA_CONCURRENT_ACCESS;

	    if(pixmap)
		mode->flags |= DGA_PIXMAP_AVAILABLE;
	    if(!pMQ->NoAccel)
            {
		mode->flags |= DGA_FILL_RECT | DGA_BLIT_RECT;
		//HSU: No transparent blt for now
                //mode->flags |= DGA_BLIT_RECT_TRANS;
	    }
            //HSU: No support for DBLSCAN and INTERLACE modes
	    //if(pMode->Flags & V_DBLSCAN)
	    //	mode->flags |= DGA_DOUBLESCAN;
	    //if(pMode->Flags & V_INTERLACE)
	    //  mode->flags |= DGA_INTERLACED;
	    mode->byteOrder = pScrn->imageByteOrder;
	    mode->depth = depth;
	    mode->bitsPerPixel = bitsPerPixel;
	    mode->red_mask = red;
	    mode->green_mask = green;
	    mode->blue_mask = blue;
	    mode->visualClass = visualClass;
	    mode->viewportWidth = pMode->HDisplay;
	    mode->viewportHeight = pMode->VDisplay;
            //MQ200 always use 128 bit alignment for Base address.
            //If 24bpp, it's 0 for xviewportstep
	    mode->xViewportStep = (Bpp == 3) ? 0 : (16 / Bpp);
	    mode->yViewportStep = 1;
	    mode->viewportFlags = DGA_FLIP_RETRACE;
	    mode->offset = 0;  /* gonna need to fix that */
	    mode->address = pMQ->FbStart;
	    mode->bytesPerScanline = pitch * Bpp;
	    mode->imageWidth = pitch;
	    mode->imageHeight =  pMQ->FbUsableSize / mode->bytesPerScanline; 
	    mode->pixmapWidth = pitch;
	    mode->pixmapHeight = (min(pMQ->FbUsableSize, MQ200_FB_SIZE)) / 
				mode->bytesPerScanline;
	    mode->maxViewportX = mode->imageWidth - mode->viewportWidth;
	    mode->maxViewportY = (pMQ->FbUsableSize / mode->bytesPerScanline) -
				mode->viewportHeight;

	    (*num)++;
	}

	pMode = pMode->next;
	if(pMode == firstMode)
	   break;
    }

    if(secondPitch) {
	secondPitch = 0;
	goto SECOND_PASS;
    }

    return modes;
}


Bool
MQDGAInit(ScreenPtr pScreen)
{
   ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
   MQPtr pMQ = MQPTR(pScrn);
   DGAModePtr modes = NULL;
   int num = 0;

   /* 8 */
   modes = MQSetupDGAMode (pScrn, modes, &num, 8, 8, 
		(pScrn->bitsPerPixel == 8),
		(pScrn->bitsPerPixel != 8) ? 0 : pScrn->displayWidth,
		0, 0, 0, PseudoColor);

   /* 16 */
   modes = MQSetupDGAMode (pScrn, modes, &num, 16, 16, 
		(pScrn->bitsPerPixel == 16),
		(pScrn->depth != 16) ? 0 : pScrn->displayWidth,
		0xf800, 0x07e0, 0x001f, TrueColor);

   modes = MQSetupDGAMode (pScrn, modes, &num, 16, 16, 
		(pScrn->bitsPerPixel == 16),
		(pScrn->depth != 16) ? 0 : pScrn->displayWidth,
		0xf800, 0x07e0, 0x001f, DirectColor);

#ifdef CHECK_24BPP      //No 24bpp support yet
   /* 24 */
   modes = MQSetupDGAMode (pScrn, modes, &num, 24, 24, 
		(pScrn->bitsPerPixel == 24),
		(pScrn->bitsPerPixel != 24) ? 0 : pScrn->displayWidth,
		0xff0000, 0x00ff00, 0x0000ff, TrueColor);

   modes = MQSetupDGAMode (pScrn, modes, &num, 24, 24, 
		(pScrn->bitsPerPixel == 24),
		(pScrn->bitsPerPixel != 24) ? 0 : pScrn->displayWidth,
		0xff0000, 0x00ff00, 0x0000ff, DirectColor);
#endif

   /* 32 */
   modes = MQSetupDGAMode (pScrn, modes, &num, 32, 24, 
		(pScrn->bitsPerPixel == 32),
		(pScrn->bitsPerPixel != 32) ? 0 : pScrn->displayWidth,
		0xff0000, 0x00ff00, 0x0000ff, TrueColor);

   modes = MQSetupDGAMode (pScrn, modes, &num, 32, 24, 
		(pScrn->bitsPerPixel == 32),
		(pScrn->bitsPerPixel != 32) ? 0 : pScrn->displayWidth,
		0xff0000, 0x00ff00, 0x0000ff, DirectColor);

   pMQ->numDGAModes = num;
   pMQ->DGAModes = modes;

   return DGAInit(pScreen, &MQ_DGAFuncs, modes, num);  
}


static int 
BitsSet(unsigned long data)
{
   unsigned long mask;
   int set = 0;

   for(mask = 1; mask; mask <<= 1)
        if(mask & data) set++;   

   return set;
}

static Bool
MQ_SetMode(
   ScrnInfoPtr pScrn,
   DGAModePtr pMode
)
{
   static MQFBLayout SavedLayouts[MAXSCREENS];
   int index = pScrn->pScreen->myNum;

   MQPtr pMQ = MQPTR(pScrn);

   if(!pMode)
   { /* restore the original mode */
        if(pMQ->DGAactive)
            memcpy(&pMQ->CurrentLayout, &SavedLayouts[index],
                sizeof(MQFBLayout));

        pScrn->currentMode = pMQ->CurrentLayout.mode;
        MQSwitchMode(index, pScrn->currentMode, 0);
        MQAdjustFrame(index, pScrn->frameX0, pScrn->frameY0, 0);
        pMQ->DGAactive = FALSE;
   }
   else
   {
        if(!pMQ->DGAactive)
        {  /* save the old parameters */
	    memcpy(&SavedLayouts[index], &pMQ->CurrentLayout,
                sizeof(MQFBLayout));
	    pMQ->DGAactive = TRUE;
        }
        /* update CurrentLayout */
        pMQ->CurrentLayout.bitsPerPixel = pMode->bitsPerPixel;
        pMQ->CurrentLayout.depth = pMode->depth;
        pMQ->CurrentLayout.displayWidth = pMode->bytesPerScanline / 
                              (pMode->bitsPerPixel >> 3);
        pMQ->CurrentLayout.weight.red = BitsSet(pMode->red_mask);
        pMQ->CurrentLayout.weight.green = BitsSet(pMode->green_mask);
        pMQ->CurrentLayout.weight.blue = BitsSet(pMode->blue_mask);
        /* MQModeInit() will set the mode field */

        MQSwitchMode(index, pMode->mode, 0);
   }
   
   return TRUE;
}



static int  
MQ_GetViewport(
    ScrnInfoPtr pScrn
)
{
    MQPtr pMQ = MQPTR(pScrn);

    return pMQ->DGAViewportStatus;
}

static void 
MQ_SetViewport(
    ScrnInfoPtr pScrn, 
    int x, int y, 
    int flags
)
{
    MQPtr pMQ = MQPTR(pScrn);

    MQAdjustFrame(pScrn->pScreen->myNum, x, y, flags);
    pMQ->DGAViewportStatus = 0;  /* MQAdjustFrame loops until finished */
}

static void 
MQ_FillRect (
    ScrnInfoPtr pScrn, 
    int x, int y, int w, int h, 
    unsigned long color
)
{
    MQPtr pMQ = MQPTR(pScrn);

    if(!pMQ->AccelInfoRec)
        return;

    switch(pMQ->CurrentLayout.bitsPerPixel)
    {
        case 8:
        case 16:
        case 32:
	    MQSetupForSolidFill(pScrn, color, GXcopy, ~0);
            MQSubsequentSolidFillRect(pScrn, x, y, w, h);
	    break;
        case 24:
            /* MQ24SetupForSolidFill(pScrn, color, GXcopy, ~0);
               MQ24SubsequentSolidFillRect(pScrn, x, y, w, h); */
	break;
    }

    pMQ->AccelInfoRec->NeedToSync = TRUE;;
}

static void 
MQ_BlitRect(
   ScrnInfoPtr pScrn, 
   int srcx, int srcy, 
   int w, int h, 
   int dstx, int dsty
)
{
    MQPtr pMQ = MQPTR(pScrn);
    int xdir = ((srcx < dstx) && (srcy == dsty)) ? -1 : 1;
    int ydir = (srcy < dsty) ? -1 : 1;

    if(!pMQ->AccelInfoRec)
        return;

    switch(pMQ->CurrentLayout.bitsPerPixel)
    {
        case 8:
        case 16:
        case 32:
	    MQSetupForScreenToScreenCopy(pScrn, xdir, ydir, GXcopy, ~0, -1);
            MQSubsequentScreenToScreenCopy(pScrn, srcx, srcy, dstx, dsty,
                                           w, h);
	    break;
        case 24:
            /*
	    MQ24SetupForScreenToScreenCopy(pScrn, xdir, ydir, GXcopy, ~0, -1);
            MQ24SubsequentScreenToScreenCopy(pScrn, srcx, srcy, dstx, dsty,
                                           w, h);
            */
	    break;
    }
 
    pMQ->AccelInfoRec->NeedToSync = TRUE;;
}


#if 0
static void 
MQ_BlitTransRect(
   ScrnInfoPtr pScrn, 
   int srcx, int srcy, 
   int w, int h, 
   int dstx, int dsty,
   unsigned long color
)
{
    MQPtr pMQ = MQPTR(pScrn);
    int xdir = ((srcx < dstx) && (srcy == dsty)) ? -1 : 1;
    int ydir = (srcy < dsty) ? -1 : 1;

    if(!pMQ->AccelInfoRec) return;
    if(pMQ->CurrentLayout.bitsPerPixel == 24)
        return;

    pMQ->DrawTransparent = TRUE;        //this flag not exist in our MQ struc
    switch(pMQ->CurrentLayout.bitsPerPixel)
    {
        case 8:
        case 16:
        case 32:
	    MQSetupForScreenToScreenCopy(pScrn, xdir, ydir, GXcopy, ~0, color);
            pMQ->DrawTransparent = FALSE;
            MQSubsequentScreenToScreenCopy(pScrn, srcx, srcy, dstx, dsty, w, h);
	    break;
        case 24:
            /*
	    MQ24SetupForScreenToScreenCopy(pScrn, xdir, ydir, GXcopy, ~0, -1);
            pMQ->DrawTransparent = FALSE;
            MQ24SubsequentScreenToScreenCopy(pScrn, srcx, srcy, dstx, dsty,
                                           w, h);
            */
	    break;
    }
    pMQ->AccelInfoRec->NeedToSync = TRUE;;
}
#endif


static Bool 
MQ_OpenFramebuffer(
   ScrnInfoPtr pScrn, 
   char **name,
   unsigned char **mem,
   int *size,
   int *offset,
   int *flags
)
{
    MQPtr pMQ = MQPTR(pScrn);

    *name = NULL; 		/* no special device */
    *mem = (unsigned char*)pMQ->FbAddress;
    *size = pMQ->FbMapSize;
    *offset = 0;
    *flags = DGA_NEED_ROOT;

    return TRUE;
}
