/* $XConsortium: MQ_driver.c $ */

/*
 This MediaQ driver is ported from MQ driver.
*/


/* All drivers should typically include these */
#include "xf86.h"
#include "xf86_OSproc.h"
#include "xf86Resources.h"

/* All drivers need this */
#include "xf86_ansic.h"

/*#include "compiler.h"*/

/* Drivers for PCI hardware need this */
#include "xf86PciInfo.h"

/* Drivers that need to access the PCI config space directly need this */
#include "xf86Pci.h"

/* All drivers initialising the SW cursor need this */
#include "mipointer.h"

/* All drivers implementing backing store need this */
#include "mibstore.h"

#include "micmap.h"

#include "xf86RAC.h"



/*
 * If using cfb, cfb.h is required.  Select the others for the bpp values
 * the driver supports.
 */
#define PSZ 8	/* needed for cfb.h */
#include "cfb.h"
#undef PSZ
#include "cfb16.h"
#include "cfb24.h"
#include "cfb32.h"

#include "mq.h"
#include "mqplat.h"
#include "mq2hw.h"
#include "mqproto.h"

#include "xaa.h"
#include "xf86cmap.h"
#include "shadowfb.h"
#include "fbdevhw.h"


/*
 * Forward definitions for the functions that make up the driver.
 */

/* Mandatory functions */
static OptionInfoPtr	MQAvailableOptions(int chipid, int busid);
static void	MQIdentify(int flags);
static Bool	MQProbe(DriverPtr drv, int flags);
static Bool	MQPreInit(ScrnInfoPtr pScrn, int flags);
static Bool	MQScreenInit(int Index, ScreenPtr pScreen, int argc,
			      char **argv);
static Bool	MQEnterVT(int scrnIndex, int flags);
static Bool	MQEnterVTFBDev(int scrnIndex, int flags);
static void	MQLeaveVT(int scrnIndex, int flags);
static Bool	MQCloseScreen(int scrnIndex, ScreenPtr pScreen);
static Bool	MQSaveScreen(ScreenPtr pScreen, int mode);

/* Optional functions */
static void	MQFreeScreen(int scrnIndex, int flags);
static int	MQValidMode(int scrnIndex, DisplayModePtr mode, Bool verbose,
			     int flags);
#ifdef DPMSExtension
static void	MQDisplayPowerManagementSet(ScrnInfoPtr pScrn,
					     int PowerManagementMode,
					     int flags);
#endif

/* Internally used functions */
static Bool	MQMapMem(ScrnInfoPtr pScrn);
static Bool	MQUnmapMem(ScrnInfoPtr pScrn);
static void	MQSave(ScrnInfoPtr pScrn);
static void	MQRestore(ScrnInfoPtr pScrn);
static Bool	MQModeInit(ScrnInfoPtr pScrn, DisplayModePtr mode);


#define VERSION 4000
#define MQ_NAME "MEDIAQ"
#define MQ_DRIVER_NAME "mediaq"
#define MQ_MAJOR_VERSION 1
#define MQ_MINOR_VERSION 0
#define MQ_PATCHLEVEL 0

/*
 * This contains the functions needed by the server after loading the driver
 * module.  It must be supplied, and gets passed back by the SetupProc
 * function in the dynamic case.  In the static case, a reference to this
 * is compiled in, and this requires that the name of this DriverRec be
 * an upper-case version of the driver name.
 */

DriverRec MEDIAQ =
{
    VERSION,
    MQ_DRIVER_NAME,
#if 0
    "Driver for MediaQ chipsets",
#endif
    MQIdentify,
    MQProbe,
    MQAvailableOptions,
    NULL,
    0
};

static SymTabRec MQChipsets[] =
{
    { PCI_CHIP_MQ200,   "MQ200" },
    { -1,		 NULL }
};

static PciChipsets MQPciChipsets[] =
{
    { PCI_CHIP_MQ200,	PCI_CHIP_MQ200,	RES_UNDEFINED},
    { -1,		-1,		RES_UNDEFINED }
};

/* The options supported by the MediaQ Driver */
typedef enum
{
    OPTION_SW_CURSOR,
    OPTION_HW_CURSOR,
    OPTION_NOACCEL,
    OPTION_SHADOW_FB,
    OPTION_FBDEV,
    OPTION_ROTATE,
    OPTION_MQFLAG,
    OPTION_MQ_OEMPANELX,
    OPTION_MQ_OEMPANELY,
    OPTION_MQ_OEMFREQ,
    OPTION_MQ_OEMFPCTRL,
    OPTION_MQ_OEMFPPINCTRL,
    OPTION_MQ_OEMSTNCTRL,
    OPTION_MQ_OEMHD,
    OPTION_MQ_OEMVD,
    OPTION_MQ_OEMHS,
    OPTION_MQ_OEMVS,
    OPTION_MQ_OEMCRTC,
    OPTION_MQ_OEMPLL
} MQOpts;

static OptionInfoRec MQOptions[] =
{
    { OPTION_SW_CURSOR,		"SWcursor",	OPTV_BOOLEAN,	{0}, FALSE },
    { OPTION_HW_CURSOR,		"HWcursor",	OPTV_BOOLEAN,	{0}, FALSE },
    { OPTION_NOACCEL,		"NoAccel",	OPTV_BOOLEAN,	{0}, FALSE },
    { OPTION_SHADOW_FB,		"ShadowFB",	OPTV_BOOLEAN,	{0}, FALSE },
    { OPTION_FBDEV,		"UseFBDev",	OPTV_BOOLEAN,	{0}, FALSE },
    { OPTION_ROTATE,		"Rotate",	OPTV_ANYSTR,	{0}, FALSE },
    { OPTION_MQFLAG,		"Flag",	        OPTV_ANYSTR,	{0}, FALSE },
    { OPTION_MQ_OEMPANELX,	"OEM_PanelX",	OPTV_INTEGER,	{0}, FALSE },
    { OPTION_MQ_OEMPANELY,	"OEM_PanelY",	OPTV_INTEGER,	{0}, FALSE },
    { OPTION_MQ_OEMFREQ,	"OEM_Freq",	OPTV_INTEGER,	{0}, FALSE },
    { OPTION_MQ_OEMFPCTRL,	"OEM_FPControl",   OPTV_ANYSTR,	{0}, FALSE },
    { OPTION_MQ_OEMFPPINCTRL,	"OEM_FPPinControl",OPTV_ANYSTR,	{0}, FALSE },
    { OPTION_MQ_OEMSTNCTRL,	"OEM_STNControl",  OPTV_ANYSTR,	{0}, FALSE },
    { OPTION_MQ_OEMHD,		"OEM_HD",	OPTV_ANYSTR,	{0}, FALSE },
    { OPTION_MQ_OEMVD,		"OEM_VD",	OPTV_ANYSTR,	{0}, FALSE },
    { OPTION_MQ_OEMHS,		"OEM_HS",	OPTV_ANYSTR,	{0}, FALSE },
    { OPTION_MQ_OEMVS,		"OEM_VS",	OPTV_ANYSTR,	{0}, FALSE },
    { OPTION_MQ_OEMCRTC,	"OEM_CRTC",	OPTV_ANYSTR,	{0}, FALSE },
    { OPTION_MQ_OEMPLL,		"OEM_PLL",	OPTV_ANYSTR,	{0}, FALSE },
    { -1,			NULL,		OPTV_NONE,	{0}, FALSE }
};


/*
 * List of symbols from other modules that this module references.  This
 * list is used to tell the loader that it is OK for symbols here to be
 * unresolved providing that it hasn't been told that they haven't been
 * told that they are essential via a call to xf86LoaderReqSymbols() or
 * xf86LoaderReqSymLists().  The purpose is this is to avoid warnings about
 * unresolved symbols that are not required.
 */


static const char *cfbSymbols[] =
{
    "cfbScreenInit",
    "cfb16ScreenInit",
    "cfb24ScreenInit",
    "cfb32ScreenInit",
    NULL
};

static const char *xaaSymbols[] =
{
    "XAADestroyInfoRec",
    "XAACreateInfoRec",
    "XAAInit",
    "XAAScreenIndex",
    NULL
};

static const char *ramdacSymbols[] =
{
    "xf86InitCursor",
    "xf86CreateCursorInfoRec",
    "xf86DestroyCursorInfoRec",
    NULL
};

static const char *shadowSymbols[] =
{
    "ShadowFBInit",
    NULL
};

static const char *fbdevHWSymbols[] =
{
    "fbdevHWInit",
    "fbdevHWUseBuildinMode",

    "fbdevHWGetDepth",
    "fbdevHWGetVidmem",

    /* colormap */
    "fbdevHWLoadPalette",

    /* ScrnInfo hooks */
    "fbdevHWSwitchMode",
    "fbdevHWAdjustFrame",
    "fbdevHWEnterVT",
    "fbdevHWLeaveVT",
    "fbdevHWValidMode",
    "fbdevHWRestore",
    "fbdevHWModeInit",
    "fbdevHWSave",

    "fbdevHWUnmapMMIO",
    "fbdevHWUnmapVidmem",
    "fbdevHWMapMMIO",
    "fbdevHWMapVidmem",
	
    NULL
};

#ifdef XFree86LOADER

static MODULESETUPPROTO(mqSetup);

static XF86ModuleVersionInfo mqVersRec =
{
    "mediaq",
    MODULEVENDORSTRING,
    MODINFOSTRING1,
    MODINFOSTRING2,
    XF86_VERSION_CURRENT,
    MQ_MAJOR_VERSION, MQ_MINOR_VERSION, MQ_PATCHLEVEL,
    ABI_CLASS_VIDEODRV,
    ABI_VIDEODRV_VERSION,
    MOD_CLASS_VIDEODRV,
    {0,0,0,0}
};

XF86ModuleData mediaqModuleData = { &mqVersRec, mqSetup, NULL };

static pointer
mqSetup(pointer module, pointer opts, int *errmaj, int *errmin)
{
    static Bool setupDone = FALSE;

    /* This module should be loaded only once, but check to be sure. */

    if (!setupDone)
    {
	setupDone = TRUE;
	xf86AddDriver(&MEDIAQ, module, 0);

	/*
	 * Modules that this driver always requires may be loaded here
	 * by calling LoadSubModule().
	 */

	/*
	 * Tell the loader about symbols from other modules that this module
	 * might refer to.
	 */
	LoaderRefSymLists(cfbSymbols, xaaSymbols, 
			  ramdacSymbols,
			  shadowSymbols,
			  fbdevHWSymbols,
			  NULL);

	/*
	 * The return value must be non-NULL on success even though there
	 * is no TearDownProc.
	 */
	return (pointer)1;
    }
    else
    {
	if (errmaj)
	    *errmaj = LDR_ONCEONLY;
	return NULL;
    }
}

#endif /* XFree86LOADER */

/*
 * This is intentionally screen-independent.  It indicates the binding
 * choice made in the first PreInit.
 */
static int pix24bpp = 0;

/* 
 * ramdac info structure initialization
 */
 /*
static MQRamdacRec DacInit =
{
    FALSE, 0, 0, 0, 0, NULL, NULL, NULL, NULL, NULL, NULL
}; 
*/

static Bool
MQGetRec(ScrnInfoPtr pScrn)
{
    /*
     * Allocate an MQRec, and hook it into pScrn->driverPrivate.
     * pScrn->driverPrivate is initialised to NULL, so we can check if
     * the allocation has already been done.
     */
    if (pScrn->driverPrivate != NULL)
	return TRUE;

    pScrn->driverPrivate = xnfcalloc(sizeof(MQRec), 1);
    /* Initialise it */

    //MQPTR(pScrn)->Dac = DacInit;
    return TRUE;
}

static void
MQFreeRec(ScrnInfoPtr pScrn)
{
    if (pScrn->driverPrivate == NULL)
	return;
    xfree(pScrn->driverPrivate);
    pScrn->driverPrivate = NULL;
}

static
OptionInfoPtr
MQAvailableOptions(int chipid, int busid)
{
    return MQOptions;
}

/* Mandatory */
static void
MQIdentify(int flags)
{
    xf86PrintChipsets(MQ_NAME, "driver for MediaQ chipsets", MQChipsets);
}


/* Mandatory */
static Bool
MQProbe(DriverPtr drv, int flags)
{
    int i;
    GDevPtr *devSections = NULL;
    int *usedChips;
    int numDevSections;
    int numUsed = 0;
    Bool foundScreen = FALSE;

    /*
     * The aim here is to find all cards that this driver can handle,
     * and for the ones not already claimed by another driver, claim the
     * slot, and allocate a ScrnInfoRec.
     *
     * This should be a minimal probe, and it should under no circumstances
     * change the state of the hardware.  Because a device is found, don't
     * assume that it will be used.  Don't do any initialisations other than
     * the required ScrnInfoRec initialisations.  Don't allocate any new
     * data structures.
     */

    /*
     * Check if there has been a chipset override in the config file.
     * For this we must find out if there is an active device section which
     * is relevant, i.e., which has no driver specified or has THIS driver
     * specified.
     */

    if ((numDevSections = xf86MatchDevice(MQ_DRIVER_NAME,
					  &devSections)) <= 0)
    {
	/*
	 * There's no matching device section in the config file, so quit
	 * now.
	 */
	return FALSE;
    }

    /*
     * We need to probe the hardware first.  We then need to see how this
     * fits in with what is given in the config file, and allow the config
     * file info to override any contradictions.
     */

#ifdef MQ_X86
    #ifdef MQ_PCI
    /*
     * All of the cards this driver supports are PCI, so the "probing" just
     * amounts to checking the PCI data that the server has already collected.
     */
    if (xf86GetPciVideoInfo() == NULL)
    {
	/*
	 * We won't let anything in the config file override finding no
	 * PCI video cards at all.  This seems reasonable now, but we'll see.
	 */
	return FALSE;
    }

    numUsed = xf86MatchPciInstances(MQ_DRIVER_NAME, PCI_VENDOR_MEDIAQ,
			MQChipsets, MQPciChipsets, devSections,
			numDevSections, drv, &usedChips);
    /* Free it since we don't need that list after this */

    if (devSections)
	xfree(devSections);
    devSections = NULL;
    #endif	//MQ_PCI
#endif 		//MQ_X86
#ifdef MQ_SA1110
#endif	//MQ_SA1110
#ifdef	MQ_MIPS_NEC
#endif	//MQ_MIPS_NEC
#ifdef	MQ_MIPS_TOSHIBA
#endif	//MQ_MIPS_TOSHIBA
#ifdef	MQ_SHx
#endif	//MQ_SHx

    if (numUsed <= 0)
	return FALSE;

    for (i = 0; i < numUsed; i++)
    {
	ScrnInfoPtr pScrn;
	
	/* Allocate a ScrnInfoRec and claim the slot */
	pScrn = xf86AllocateScreen(drv, 0);
	
	/* Fill in what we can of the ScrnInfoRec */
	pScrn->driverVersion	= VERSION;
	pScrn->driverName	= MQ_DRIVER_NAME;
	pScrn->name		= MQ_NAME;
	pScrn->Probe		= MQProbe;
	pScrn->PreInit		= MQPreInit;
	pScrn->ScreenInit	= MQScreenInit;
	pScrn->SwitchMode	= MQSwitchMode;
	pScrn->AdjustFrame	= MQAdjustFrame;
	pScrn->EnterVT		= MQEnterVT;
	pScrn->LeaveVT		= MQLeaveVT;
	pScrn->FreeScreen	= MQFreeScreen;
	pScrn->ValidMode	= MQValidMode;
	foundScreen = TRUE;
	xf86ConfigActivePciEntity(pScrn, usedChips[i], MQPciChipsets, NULL,
				      NULL, NULL, NULL, NULL);
    	xfree(usedChips);
    }
    return foundScreen;
}

/* Mandatory */
static Bool
MQPreInit(ScrnInfoPtr pScrn, int flags)
{
    MQPtr pMQ;
    MessageType from;
    int i;
    char *mod = NULL;
    const char *reqSym = NULL;
    const char *s;
    int flags24;
    ULONG paneltype;
    PFPDATA_CONTROL pFP;

    /*
     * Note: This function is only called once at server startup, and
     * not at the start of each server generation.  This means that
     * only things that are persistent across server generations can
     * be initialised here.  xf86Screens[] is (pScrn is a pointer to one
     * of these).  Privates allocated using xf86AllocateScrnInfoPrivateIndex()  
     * are too, and should be used for data that must persist across
     * server generations.
     *
     * Per-generation data should be allocated with
     * AllocateScreenPrivateIndex() from the ScreenInit() function.
     */

    /* Check the number of entities, and fail if it isn't one. */
    if (pScrn->numEntities != 1)
	return FALSE;

    /* Allocate the MQRec driverPrivate */
    if (!MQGetRec(pScrn))
    {
	return FALSE;
    }
    pMQ = MQPTR(pScrn);

    /* Get the entity */
    pMQ->pEnt = xf86GetEntityInfo(pScrn->entityList[0]);

#ifdef MQ_PCI
    /* make sure it is PCI. */
    if (pMQ->pEnt->location.type != BUS_PCI)
	return FALSE;
	
    /* Find the PCI info for this screen */
    pMQ->PciInfo = xf86GetPciInfoForEntity(pMQ->pEnt->index);
    pMQ->PciTag = pciTag(pMQ->PciInfo->bus, pMQ->PciInfo->device,
			  pMQ->PciInfo->func);
    pMQ->Primary = xf86IsPrimaryPci(pMQ->PciInfo);
#endif	//MQ_PCI

    /* Set pScrn->monitor */
    pScrn->monitor = pScrn->confScreen->monitor;

    /*
     * The first thing we should figure out is the depth, bpp, etc.
     * Our default depth is 8, so pass it to the helper function.
     * We support both 24bpp and 32bpp layouts, so indicate that.
     */

    flags24 = Support24bppFb | Support32bppFb;

    if (!xf86SetDepthBpp(pScrn, 8, 8, 8, flags24))
    {
	return FALSE;
    }
    else
    {
	/* Check that the returned depth is one we support */
	switch (pScrn->depth)
	{
	    case 8:
	    case 16:
	    case 24:
	    case 32:
		break; 		/* OK */
	    default:
		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
		       "Given depth (%d) is not supported by this driver\n",
		       pScrn->depth);
		return FALSE;
	}
    }
    xf86PrintDepthBpp(pScrn);

    /* Get the depth24 pixmap format */
    if (pScrn->depth == 24 && pix24bpp == 0)
	pix24bpp = xf86GetBppFromDepth(pScrn, 24);

    /*
     * This must happen after pScrn->display has been set because
     * xf86SetWeight references it.
     */
    if (pScrn->depth > 8)
    {
	/* The defaults are OK for us */
	rgb zeros = {0, 0, 0};

	if (!xf86SetWeight(pScrn, zeros, zeros))
	{
	    return FALSE;
	}
	else
	{
	    /* XXX check that weight returned is supported */
            ;
        }
    }

    if (!xf86SetDefaultVisual(pScrn, -1))
	return FALSE;

    /* Collect all of the relevant option flags (fill in pScrn->options) */
    xf86CollectOptions(pScrn, NULL);

    /* Process the options */
    xf86ProcessOptions(pScrn->scrnIndex, pScrn->options, MQOptions);

    /* Set the bits per RGB for 8bpp mode */
    if (pScrn->depth == 8)
	pScrn->rgbBits = 8;

    /*
     * Set the Chipset and ChipRev, allowing config file entries to
     * override.
     */
    if (pMQ->pEnt->device->chipset && *pMQ->pEnt->device->chipset)
    {
	pScrn->chipset = pMQ->pEnt->device->chipset;
        pMQ->Chipset = xf86StringToToken(MQChipsets, pScrn->chipset);
        from = X_CONFIG;
    }
    else if (pMQ->pEnt->device->chipID >= 0)
    {
	pMQ->Chipset = pMQ->pEnt->device->chipID;
	pScrn->chipset = (char *)xf86TokenToString(MQChipsets, pMQ->Chipset);
	from = X_CONFIG;
	xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "ChipID override: 0x%04X\n",
		   pMQ->Chipset);
    }
    else
    {
    #ifdef MQ_PCI
	from = X_PROBED;
	pMQ->Chipset = pMQ->PciInfo->chipType;
	pScrn->chipset = (char *)xf86TokenToString(MQChipsets, pMQ->Chipset);
    #endif	//MQ_PCI
    }

    if (pMQ->pEnt->device->chipRev >= 0)
    {
	pMQ->ChipRev = pMQ->pEnt->device->chipRev;
	xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "ChipRev override: %d\n",
		   pMQ->ChipRev);
    }
    else
    {
    #ifdef MQ_PCI
	pMQ->ChipRev = pMQ->PciInfo->chipRev;
    #endif	//MQ_PCI
    }

    /*
     * This shouldn't happen because such problems should be caught in
     * MQProbe(), but check it just in case.
     */
    if (pScrn->chipset == NULL)
    {
	xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
		   "ChipID 0x%04X is not recognised\n", pMQ->Chipset);
	return FALSE;
    }
    if (pMQ->Chipset < 0)
    {
	xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
		   "Chipset \"%s\" is not recognised\n", pScrn->chipset);
	return FALSE;
    }

    xf86DrvMsg(pScrn->scrnIndex, from, "Chipset: \"%s\"\n", pScrn->chipset);

    from = X_DEFAULT;

    /* MediaQ - begin specifc options */
    pMQ->HWCursor = TRUE;
    pMQ->NoAccel = FALSE;       //Default is to turn on acceleration
                                //unless otherwise
    //pMQ->MQFlag = 0x130004;	//QV Same Image, panel type 4
    //pMQ->MQFlag = 0x30004;	//Simul LCD/CRT, panel type 4
    pMQ->MQFlag = 0x0;		//Default flag to 0.  Panel setup will
    				//come from XF86Config file

    MQGetOptValInteger(MQOptions, OPTION_MQFLAG, &(pMQ->MQFlag));
    //if ( MQGetOptValInteger(MQOptions, OPTION_MQFLAG, &(pMQ->MQFlag)) )
	//xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "MQFlag: 0x%x\n", pMQ->MQFlag);

    /* if ((s = xf86GetOptValString(MQOptions, OPTION_MQFLAG)))
    {
 	pMQ->MQFlag = strtoul(s, NULL, 16);
	xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "MQFlag: 0x%x\n", pMQ->MQFlag);
    }
    if ( (s && !(pMQ->MQFlag & PANEL_TYPE_MASK)) || !s )
    */
    if ( !(pMQ->MQFlag & PANEL_TYPE_MASK) )
    {
    	//Oem specific panel.  Need to fill 0th structure.

	#define	DEFAULT_X		640
	#define	DEFAULT_Y		480
	#define	DEFAULT_FREQ		60
	#define	DEFAULT_BPP		8

	PDISPLAY_TIMING pt = &(TimingParam[0]);
	BOOL	found;
	int	val;

	pFP = &(fpControlData[0]);
	pt->x = xf86GetOptValInteger(MQOptions, OPTION_MQ_OEMPANELX, &val) ?
		 (( val == -1 ) ? DEFAULT_X : val) : DEFAULT_X;
	pt->y = xf86GetOptValInteger(MQOptions, OPTION_MQ_OEMPANELY, &val) ?
		 (( val == -1 ) ? DEFAULT_Y : val) : DEFAULT_Y;
	pt->refresh = xf86GetOptValInteger(MQOptions, OPTION_MQ_OEMFREQ, &val) ?
		 (( val == -1 ) ? DEFAULT_FREQ : val) : DEFAULT_FREQ;
 	ErrorF("Token x: %d\nToken y: %d\nToken freq: %d\n",
		pt->x, pt->y, pt->refresh);
#if 1
	found = MQGetOptValULONG(MQOptions, OPTION_MQ_OEMFPCTRL,
							&(pFP->fpControl))
		&& MQGetOptValULONG(MQOptions, OPTION_MQ_OEMFPPINCTRL,
							&(pFP->fpPinControl))
		&& MQGetOptValULONG(MQOptions, OPTION_MQ_OEMSTNCTRL,
							&(pFP->stnControl))
		&& MQGetOptValULONG(MQOptions, OPTION_MQ_OEMHD, &(pt->hd))
		&& MQGetOptValULONG(MQOptions, OPTION_MQ_OEMVD, &(pt->vd))
		&& MQGetOptValULONG(MQOptions, OPTION_MQ_OEMHS, &(pt->hs))
		&& MQGetOptValULONG(MQOptions, OPTION_MQ_OEMVS, &(pt->vs))
		&& MQGetOptValULONG(MQOptions, OPTION_MQ_OEMCRTC, &(pt->crtc))
		&& MQGetOptValULONG(MQOptions, OPTION_MQ_OEMPLL, &(pt->pll));
#else
	found = xf86GetOptValInteger(MQOptions, OPTION_MQ_OEMFPCTRL,
			(int *)&(pFP->fpControl))
		&& xf86GetOptValInteger(MQOptions, OPTION_MQ_OEMFPPINCTRL,
			(int *)&(pFP->fpPinControl))
		&& xf86GetOptValInteger(MQOptions, OPTION_MQ_OEMSTNCTRL,
			(int *)&(pFP->stnControl))
		&& xf86GetOptValInteger(MQOptions, OPTION_MQ_OEMHD,
			(int *)&(pt->hd))
		&& xf86GetOptValInteger(MQOptions, OPTION_MQ_OEMVD,
			(int *)&(pt->vd))
		&& xf86GetOptValInteger(MQOptions, OPTION_MQ_OEMHS,
			(int *)&(pt->hs))
		&& xf86GetOptValInteger(MQOptions, OPTION_MQ_OEMVS,
			(int *)&(pt->vs))
		&& xf86GetOptValInteger(MQOptions, OPTION_MQ_OEMCRTC,
			(int *)&(pt->crtc))
		&& xf86GetOptValInteger(MQOptions, OPTION_MQ_OEMPLL,
			(int *)&(pt->pll));
#endif
	if ( !found )
	{
	    //if ( !(pMQ->MQFlag & CRT_ON) )
	    {
	    	xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
		    "Insufficient Panel information for OEM display type!.");
		MQFreeRec(pScrn);
		return FALSE;
	    }
	    /*
	    //Keep CRT_ON only if it was specified.
	    pMQ->MQFlag &= ~(LCD_ON & LARGE_DESKTOP & SAME_IMAGE & USE_2GCs);
	    for( val = 1; val < MAX_MQMODE; val++ )
	    {
	    	if ( pt[val].x == pt.x
		     && pt[val].y == pt.y
		     && pt[va].freq == pt.freq )
		{
		    pt[0] = pt[1];
		    break;
		}
	    }
	    if ( val >= MAX_MQMODE )
	    {
	    	xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
		    "Incomplete Panel/CRT information for OEM display type!.");
		MQFreeRec(pScrn);
		return FALSE;
	    }
	    */
	}
	else
	{
	    pFP->x = pt->x;
	    pFP->y = pt->y;
	    pFP->freq = pt->refresh;
	}
    }

    /* MediaQ - end specifc options */

    if (xf86ReturnOptValBool(MQOptions, OPTION_NOACCEL, FALSE))
    {
	pMQ->NoAccel = TRUE;
	xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Acceleration disabled\n");
    }
    if (xf86ReturnOptValBool(MQOptions, OPTION_SHADOW_FB, FALSE))
    {
	pMQ->ShadowFB = TRUE;
	pMQ->NoAccel = TRUE;
	xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, 
		"Using \"Shadow Framebuffer\" - acceleration disabled\n");
    }
    if (xf86ReturnOptValBool(MQOptions, OPTION_FBDEV, FALSE))
    {
	pMQ->FBDev = TRUE;
	xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, "Using framebuffer device\n");
    }
    if (pMQ->FBDev)
    {
        ErrorF("FBDev: LoadSubModule\n");
	/* check for linux framebuffer device */
	if (!xf86LoadSubModule(pScrn, "fbdevhw"))
	    return FALSE;
        ErrorF("FBDev: LoadReqSymLists\n");
	xf86LoaderReqSymLists(fbdevHWSymbols, NULL);
    #ifdef MQ_PCI
        ErrorF("FBDev: HWInit\n");
	if (!fbdevHWInit(pScrn, pMQ->PciInfo, NULL))
	    return FALSE;
        ErrorF("FBDev: After HWInit\n");
    #endif //MQ_PCI
	pScrn->SwitchMode    = MQSwitchMode;
	pScrn->AdjustFrame   = MQAdjustFrame;
	pScrn->EnterVT       = MQEnterVTFBDev;
	pScrn->LeaveVT       = MQLeaveVT;
	pScrn->ValidMode     = MQValidMode;
    /*
	pScrn->SwitchMode    = fbdevHWSwitchMode;
	pScrn->AdjustFrame   = fbdevHWAdjustFrame;
	pScrn->EnterVT       = MQEnterVTFBDev;
	pScrn->LeaveVT       = fbdevHWLeaveVT;
	pScrn->ValidMode     = fbdevHWValidMode;
     */
    }
    pMQ->Rotate = 0;
    if ((s = xf86GetOptValString(MQOptions, OPTION_ROTATE)))
    {
	if(!xf86NameCmp(s, "CW"))
	{
	    pMQ->ShadowFB = TRUE;
	    pMQ->NoAccel = TRUE;
	    pMQ->HWCursor = FALSE;
	    pMQ->Rotate = 1;
	    xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, 
		"Rotating screen clockwise - acceleration disabled\n");
	}
	else if(!xf86NameCmp(s, "CCW"))
	{
	    pMQ->ShadowFB = TRUE;
	    pMQ->NoAccel = TRUE;
	    pMQ->HWCursor = FALSE;
	    pMQ->Rotate = -1;
	    xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, 
		"Rotating screen counter clockwise - acceleration disabled\n");
	}
	else
	{
	    xf86DrvMsg(pScrn->scrnIndex, X_CONFIG, 
		"\"%s\" is not a valid value for Option \"Rotate\"\n", s);
	    xf86DrvMsg(pScrn->scrnIndex, X_INFO, 
		"Valid options are \"CW\" or \"CCW\"\n");
	}
    }

    /* Init RAMDAC */
    MQSetupFuncs(pScrn);

#ifdef MQ_PCI
    pMQ->pEnt->device->MemBase = pMQ->PciInfo->memBase[1]; 
    pMQ->pEnt->device->IOBase = pMQ->PciInfo->memBase[0];

    if (pMQ->pEnt->device->MemBase != 0)
    {
	/* Require that the config file value matches one of the PCI values. */
	if (!xf86CheckPciMemBase(pMQ->PciInfo, pMQ->pEnt->device->MemBase))
	{
	    xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
		"MemBase 0x%08lX doesn't match any PCI base register.\n",
		pMQ->pEnt->device->MemBase);
	    MQFreeRec(pScrn);
	    return FALSE;
	}
	pMQ->FbAddress = pMQ->pEnt->device->MemBase;
	from = X_CONFIG;
    }
    else
    {
	xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
		"No valid FB address in PCI config space\n");
	MQFreeRec(pScrn);
	return FALSE;
    }
    xf86DrvMsg(pScrn->scrnIndex, from, "Linear framebuffer at 0x%lX\n",
	       (unsigned long)pMQ->FbAddress);

    if (pMQ->pEnt->device->IOBase != 0)
    {
	/* Require that the config file value matches one of the PCI values. */
	if (!xf86CheckPciMemBase(pMQ->PciInfo, pMQ->pEnt->device->IOBase))
	{
	    xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
		"IOBase 0x%08lX doesn't match any PCI base register.\n",
		pMQ->pEnt->device->IOBase);
	    MQFreeRec(pScrn);
	    return FALSE;
	}
	pMQ->IOAddress = pMQ->pEnt->device->IOBase;
	from = X_CONFIG;
    }
    else
    {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
			"No valid MMIO address in PCI config space\n");
        MQFreeRec(pScrn);
	return FALSE;
    }
#endif	//MQ_PCI
    xf86DrvMsg(pScrn->scrnIndex, from, "MMIO registers at 0x%lX\n",
	       (unsigned long)pMQ->IOAddress);

    /*
     * If the driver can do gamma correction, it should call xf86SetGamma()
     * here.
     */

    {
	Gamma zeros = {0.0, 0.0, 0.0};

	if (!xf86SetGamma(pScrn, zeros)) {
	    return FALSE;
	}
    }

    /* HW bpp matches reported bpp */
    pMQ->HwBpp = pScrn->bitsPerPixel;

    /* Fill memory size */
    pMQ->FbMapSize = MQ200_FB_SIZE;
    pScrn->videoRam = MQ200_FB_SIZE / 1024;
    if ( pMQ->HWCursor )
    	pScrn->videoRam -= 1;	//-1 for hwcursor buffer

/* HSU: SkipCursor
    (*pMQ->PreInit)(pScrn);
 */

    /*
     * xf86ValidateModes will check that the mode HTotal and VTotal values
     * don't exceed the chipset's limit if pScrn->maxHValue and
     * pScrn->maxVValue are set.  Since our MQValidMode() already takes
     * care of this, we don't worry about setting them here.
     */
    {
#if 1
        //This is the plan.  We check our flag, for panel type
        //and qview configuration.  Set up the pitches, correspondingly.
        //Whatever currentmode we get from xf86ValidateModes will only
        //apply to CRT and we will use DisplayModePtr's HDisplay and
        //VDisplay for screen width and height.
        int pitches2[6];
        int pitches1[] = {640, 800, 1024, 1280, 1600, 0};
        int *pitches;
	int minPitch, maxPitch;
        int minHeight, maxHeight;
        ClockRange clockRanges;
	//DisplayModePtr	mode;

        paneltype = pMQ->MQFlag & PANEL_TYPE_MASK;
        //ErrorF("Pre-ValidateMode:%d\n", fpControlData[0].x);

	pFP = NULL;
        if (paneltype || fpControlData[0].x)
        {
            pFP = &fpControlData[paneltype];

            /* Figure out half frame buffer for DSTN panel */
            if ( (pFP->fpControl & FP_TYPE_MASK) == FP_TYPE_DSTN )
                pScrn->videoRam -= ((computehfbuffer( pFP ) + 1023) / 1024);
            if (pMQ->MQFlag & LARGE_DESKTOP)
            {
	        switch (pMQ->MQFlag & LCDCRT_POS_MASK)
                {
                    default:
                    case HORI_CRT_LCD:
                    case HORI_LCD_CRT:
                        pitches2[0] = pFP->x * 2;
                        pitches2[1] = 0;
                        minHeight = maxHeight = pFP->y;
                        break;
                    case VERT_CRT_LCD:
                    case VERT_LCD_CRT:
                        pitches2[0] = pFP->x;
                        pitches2[1] = 0;
                        minHeight = maxHeight = pFP->y * 2;
                        break;
                }
                minPitch = maxPitch = pitches2[0];
            }
            else
            {
                int     *p1;
                p1 = pitches1;
                pitches = &(pitches2[0]);
                minPitch = maxPitch = 0;
                while( *p1 != 0 )
                {
                    //if ( (*p1) <= (pFP->x * 2) )
                    if ( (*p1) >= pFP->x )
                    {
		        if ( 0 == minPitch )
			    minPitch = *p1;
                        *pitches++ = *p1;
                        maxPitch = *p1;
                    }
                    p1++;
                }
                *pitches = 0;
                minHeight = pFP->y;
		maxHeight = 2048;
            }
            pitches = &(pitches2[0]);
        }
	else
        {
            pitches = pitches1;
            minPitch = 640;
            maxPitch = 1600;
            minHeight = 480;
            maxHeight = 2048;
        }
	//computehfpanel(paneltype, pMQMMIO);
	//screensize = pDC->x * pDC->y * pDC->bpp / 8;
	//setuphfbuffer(paneltype, screensize, &(pMQ->FbUsableSize), pMQMMIO);

        clockRanges.next = NULL;
        clockRanges.minClock = 0;
        clockRanges.maxClock = 103000;         //103MHz
        clockRanges.clockIndex = -1;		/* programmable */
        clockRanges.interlaceAllowed = FALSE;
        clockRanges.doubleScanAllowed = FALSE;
 
        pScrn->progClock = TRUE;
	/*
        ErrorF("ValidateMode:%d, %d, %d, %d, %d, %d, %d, %d\n",
		minPitch, maxPitch,
                pScrn->numClocks,
                pScrn->videoRam,
                pScrn->fbFormat.scanlinePad,
                pScrn->fbFormat.bitsPerPixel,
                pScrn->display->virtualX, pScrn->display->virtualY);
	*/
	i = xf86ValidateModes(pScrn, pScrn->monitor->Modes,
			      pScrn->display->modes, &clockRanges,
			      pitches, minPitch, maxPitch,
			      128,       	//pitch incr by 128 bit
                              minHeight,	//min height
                              maxHeight,	//max height
			      pScrn->display->virtualX,
			      pScrn->display->virtualY,
			      pMQ->FbMapSize,
			      LOOKUP_BEST_REFRESH);
			      //LOOKUP_DEFAULT);
#else
        int pitches[] = {640, 800, 1024, 1152, 1280, 0};
	int minPitch = 640;
	int maxPitch = 1280;
        ClockRange clockRanges;

        clockRanges.next = NULL;
        clockRanges.minClock = 0;
        clockRanges.maxClock = 103000;         //103MHz
        clockRanges.clockIndex = -1;		/* programmable */
        clockRanges.interlaceAllowed = FALSE;
        clockRanges.doubleScanAllowed = FALSE;
 
        pScrn->progClock = TRUE;
        ErrorF("ValidateMode:%d, %d, %d, %d, %d, %d\n",
                pScrn->numClocks,
                pScrn->videoRam,
                pScrn->fbFormat.scanlinePad,
                pScrn->fbFormat.bitsPerPixel,
                pScrn->display->virtualX, pScrn->display->virtualY);
	i = xf86ValidateModes(pScrn, pScrn->monitor->Modes,
			      //pScrn->display->modes, clockRanges,
			      pScrn->display->modes, &clockRanges,
			      pitches, minPitch, maxPitch,
			      128,       //pitch incr by 128 bit
                              480,      //min height
                              1024,     //max height
			      pScrn->display->virtualX,
			      pScrn->display->virtualY,
			      pMQ->FbMapSize,
			      LOOKUP_BEST_REFRESH);
#endif

        /* HSU - ValidateMode does this */
        /* Build pScrn->modes and set each mode to MODE_OK */
        /* scrp->virtualX = virtX;
           scrp->virtualY = virtY;
           scrp->displayWidth = linePitch; */
        /* Make circular scrp->modes */
    }

    if (i == -1)
    {
        ErrorF("MQPreInit: xf86ValidateMode Fail\n");
	MQFreeRec(pScrn);
	return FALSE;
    }

    /* Prune the modes marked as invalid */
    xf86PruneDriverModes(pScrn);
    if (i == 0 || pScrn->modes == NULL)
    {
	xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "No valid modes found\n");
	MQFreeRec(pScrn);
	return FALSE;
    }

    ErrorF("MQPreInit..103:%d, %d, %d, %d, %f\n",
                pScrn->virtualX, pScrn->virtualY,
                pScrn->modes->HDisplay, pScrn->modes->VDisplay,
                pScrn->modes->VRefresh);

    /*
     * Set the CRTC parameters for all of the modes based on the type
     * of mode, and the chipset's interlace requirements.
     *
     * Calling this is required if the mode->Crtc* values are used by the
     * driver and if the driver doesn't provide code to set them.  They
     * are not pre-initialised at all.
     */

    /* Set the current mode to the first in the list */
#if 1	//HSU: Set currentMode to the one that matches flat panel
    if ( pFP )
    {
    	DisplayModePtr	newmodes, modes, omodes;
	newmodes = modes = omodes = pScrn->modes;
	do
	{
	    if ( modes->HDisplay == pFP->x && modes->VDisplay == pFP->y )
	    {
	        if ( newmodes->HDisplay == pFP->x
			&& newmodes->VDisplay == pFP->y )
		{
		    ULONG    freq = (ULONG)(modes->VRefresh + 0.5);
		    if ( freq == pFP->freq )
	    		newmodes = modes;
		}
		else
	    	    newmodes = modes;
	    }
	    modes = modes->next;
	}
	while( modes != omodes );
        pScrn->currentMode = newmodes;
	pScrn->modes = newmodes;
    }
    else
#endif
    pScrn->currentMode = pScrn->modes;

    /* Print the list of modes being used */
    xf86PrintModes(pScrn);

    /* Set display resolution */
    xf86SetDpi(pScrn, 0, 0);

    /* Load bpp-specific modules */
    switch (pScrn->bitsPerPixel)
    {
	case 8:
	    mod = "cfb";
	    reqSym = "cfbScreenInit";
	    break;
	case 16:
	    mod = "cfb16";
	    reqSym = "cfb16ScreenInit";
	    break;
	case 24:
	/*
            mod = "cfb24";
            reqSym = "cfb24ScreenInit";
	 */
	    if (pix24bpp == 24)
	    {
		mod = "cfb24";
		reqSym = "cfb24ScreenInit";
	    }
	    else
	    {
		mod = "xf24_32bpp";
		reqSym = "cfb24_32ScreenInit";
	    }
	    break;
	case 32:
            mod = "cfb32";
            reqSym = "cfb32ScreenInit";
	    break;
    }
    if (mod && xf86LoadSubModule(pScrn, mod) == NULL)
    {
	MQFreeRec(pScrn);
	return FALSE;
    }

    xf86LoaderReqSymbols(reqSym, NULL);

    /* Load XAA if needed */
    if (!pMQ->NoAccel)
    {
	if (!xf86LoadSubModule(pScrn, "xaa"))
	{
	    MQFreeRec(pScrn);
	    return FALSE;
	}
	xf86LoaderReqSymLists(xaaSymbols, NULL);
    }

    /* Load ramdac if needed */
    if (pMQ->HWCursor)
    {
	if (!xf86LoadSubModule(pScrn, "ramdac"))
        {
	    MQFreeRec(pScrn);
	    return FALSE;
	}
	xf86LoaderReqSymLists(ramdacSymbols, NULL);
    }
    else
	ErrorF("MQPreInit: No HWCursor\n");

    /* Load shadowfb if needed */
    if (pMQ->ShadowFB)
    {
	ErrorF("MQPreInit: Load ShadowFB\n");
	if (!xf86LoadSubModule(pScrn, "shadowfb"))
	{
	    MQFreeRec(pScrn);
	    return FALSE;
	}
	xf86LoaderReqSymLists(shadowSymbols, NULL);
    }

    pMQ->CurrentLayout.bitsPerPixel = pScrn->bitsPerPixel;
    pMQ->CurrentLayout.depth = pScrn->depth;
    pMQ->CurrentLayout.displayWidth = pScrn->displayWidth;
    pMQ->CurrentLayout.weight.red = pScrn->weight.red;
    pMQ->CurrentLayout.weight.green = pScrn->weight.green;
    pMQ->CurrentLayout.weight.blue = pScrn->weight.blue;
    pMQ->CurrentLayout.mode = pScrn->currentMode;

    return TRUE;
}


/*
 * Map the framebuffer and MMIO memory.
 */

static Bool
MQMapMem(ScrnInfoPtr pScrn)
{
    MQPtr pMQ;

    pMQ = MQPTR(pScrn);

    /*
     * Map IO registers to virtual address space
     */ 
#ifdef MQ_PCI
    pMQ->IOBase = xf86MapPciMem(pScrn->scrnIndex,
		 		//HSU: Don't set readsideeffect
		 		//VIDMEM_MMIO | VIDMEM_READSIDEEFFECT,
				 VIDMEM_MMIO,
				 pMQ->PciTag, pMQ->IOAddress, MQ200_MMIO_SIZE);
    if (pMQ->IOBase == NULL)
	return FALSE;
    ErrorF("MQMapMem: IOBase=%x,IOAddress=%x\n", pMQ->IOBase, pMQ->IOAddress);

    pMQ->FbBase = xf86MapPciMem(pScrn->scrnIndex, VIDMEM_FRAMEBUFFER,
				 pMQ->PciTag, pMQ->FbAddress,
				 pMQ->FbMapSize);
    if (pMQ->FbBase == NULL)
	return FALSE;
#endif	//MQ_PCI

    //HSU: Move Cursor buffer to last 1K
    //pMQ->FbStart = pMQ->FbBase + 1024; /* First KB reserved for hw cursor */
    pMQ->FbStart = pMQ->FbBase;
    ErrorF("MQMapMem: FbBase=%x,FbAddress=%x,FBSize=%x\n",
    			pMQ->FbBase, pMQ->FbAddress, pMQ->FbMapSize);
    return TRUE;
}

static Bool
MQMapMemFBDev(ScrnInfoPtr pScrn)
{
    MQPtr pMQ;

    pMQ = MQPTR(pScrn);

    pMQ->FbBase = fbdevHWMapVidmem(pScrn);
    if (pMQ->FbBase == NULL)
	return FALSE;
    ErrorF("MQMapMemFBDev: FbBase=%x\n", pMQ->FbBase);

    pMQ->IOBase = fbdevHWMapMMIO(pScrn);
    if (pMQ->IOBase == NULL)
	return FALSE;
    ErrorF("MQMapMemFBDev: IOBase=%x\n", pMQ->IOBase);

    return TRUE;
}

/*
 * Unmap the framebuffer and MMIO memory.
 */

static Bool
MQUnmapMem(ScrnInfoPtr pScrn)
{
    MQPtr pMQ;

    ErrorF("MQUnMapMem:\n");
    pMQ = MQPTR(pScrn);

#if 1
    //HSU: reset GE too during development so that GE will get reset
    {
    	ULONG	pmu;
    	pmu = pmuREAD(PM_MISC);
        if ( pmu & GE_ENABLE )
        {
	    pmuREG(PM_MISC, (pmu & ~GE_ENABLE));
            cpuREG(DC_SW_1, 0);  //break up consecutive write cycles
	    pmuREG(PM_MISC, pmu);
        }
    }
#endif

    /*
     * Unmap IO registers to virtual address space
     */ 
    xf86UnMapVidMem(pScrn->scrnIndex, (pointer)pMQ->IOBase, MQ200_MMIO_SIZE);
    pMQ->IOBase = NULL;

    xf86UnMapVidMem(pScrn->scrnIndex, (pointer)pMQ->FbBase, pMQ->FbMapSize);
    pMQ->FbBase = NULL;
    pMQ->FbStart= NULL;

    return TRUE;
}

static Bool
MQUnmapMemFBDev(ScrnInfoPtr pScrn)
{
    MQPtr pMQ;

    pMQ = MQPTR(pScrn);
    fbdevHWUnmapVidmem(pScrn);
    pMQ->FbBase = NULL;
    fbdevHWUnmapMMIO(pScrn);
    pMQ->IOBase = NULL;
    return TRUE;
}

/*
 * This function saves the video state.
 */
static void
MQSave(ScrnInfoPtr pScrn)
{
    ErrorF("MQSave\n");
}

/*
 * Initialise a new mode.  This is currently still using the old
 * "initialise struct, restore/write struct to HW" model.  That could
 * be changed.
 */

static Bool
MQModeInit(ScrnInfoPtr pScrn, DisplayModePtr mode)
{   //static DISPLAY_CONFIG dc;
    DISPLAY_CONFIG dc;
    MQPtr pMQ = MQPTR(pScrn);
    MQFBLayout *pLayout = &pMQ->CurrentLayout;

    /* ErrorF("MQModeInit: %d, %d, %d, %d, %d\n",
        		mode->HDisplay, mode->VDisplay,
			pScrn->displayWidth,
        		pLayout->bitsPerPixel, (int)(mode->VRefresh)); */

    pScrn->vtSema = TRUE;

/* HSU: SkipCursor
    if (!(*pMQ->ModeInit)(pScrn, mode))
	return FALSE;
*/

    //Package required information from different pieces into
    //our DISPLAY_CONFIG.
    dc.x = mode->HDisplay;
    dc.y = mode->VDisplay;
    dc.bpp = pLayout->bitsPerPixel;
    dc.stride = pScrn->displayWidth * dc.bpp / 8;
    dc.refresh = (int)(mode->VRefresh + 0.5);	//round vrefresh
    dc.flag = pMQ->MQFlag;
    ErrorF("MQModeInit dc:%d, %d, %d, %d, 0x%x\n",
		dc.x, dc.y, dc.bpp, dc.refresh, dc.flag);
 
    setmqmode(&dc, pMQ);

    (*pMQ->Restore)(pScrn, FALSE);

    pMQ->CurrentLayout.mode = mode;
   
    ErrorF("MQModeInit-done\n");
    return TRUE;
}

/*
 * Restore the initial (text) mode.
 */
static void 
MQRestore(ScrnInfoPtr pScrn)
{
    MQPtr pMQ = MQPTR(pScrn);

    ErrorF("MQRestore:\n");
    if (pMQ->Primary)
        (*pMQ->Restore)(pScrn, TRUE);
}


/* Mandatory */

/* This gets called at the start of each server generation */

static Bool
MQScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
{
    ScrnInfoPtr pScrn;
    MQPtr pMQ;
    //MQRamdacPtr MQdac;
    int ret = FALSE;
    VisualPtr visual;
    unsigned char *FBStart;
    int width, height, displayWidth;

    //ErrorF("MQScreenInit:\n");
    /* 
     * First get the ScrnInfoRec
     */
    pScrn = xf86Screens[pScreen->myNum];

    pMQ = MQPTR(pScrn);
    //MQdac = &pMQ->Dac;

    /* Map the MQ memory and MMIO areas */
    if (pMQ->FBDev)
    {
	if (!MQMapMemFBDev(pScrn))
	    return FALSE;
    }
    else
    {
	if (!MQMapMem(pScrn))
	    return FALSE;
    }

    if (pMQ->FBDev)
    {
	fbdevHWSave(pScrn);
	if (!fbdevHWModeInit(pScrn, pScrn->currentMode))
	    return FALSE;
    }
    else
    {
	/* Save the current state */
	MQSave(pScrn);
	/* Initialise the first mode */
	if (!MQModeInit(pScrn, pScrn->currentMode))
	    return FALSE;
    }

    /* Darken the screen for aesthetic reasons and set the viewport */
    MQSaveScreen(pScreen, SCREEN_SAVER_ON);
    pScrn->AdjustFrame(scrnIndex, pScrn->frameX0, pScrn->frameY0, 0);

    /*
     * The next step is to setup the screen's visuals, and initialise the
     * framebuffer code.  In cases where the framebuffer's default
     * choices for things like visual layouts and bits per RGB are OK,
     * this may be as simple as calling the framebuffer's ScreenInit()
     * function.  If not, the visuals will need to be setup before calling
     * a fb ScreenInit() function and fixed up after.
     *
     * For most PC hardware at depths >= 8, the defaults that cfb uses
     * are not appropriate.  In this driver, we fixup the visuals after.
     */

    /*
     * Reset the visual list.
     */
    miClearVisualTypes();

#if 0
    if (!miSetVisualTypes(pScrn->depth,
		      miGetDefaultVisualMask(pScrn->depth),
		      pScrn->rgbBits, pScrn->defaultVisual))
	return FALSE;
#else
    if (pScrn->bitsPerPixel > 8)
    {
	if (!miSetVisualTypes(pScrn->depth, TrueColorMask, pScrn->rgbBits,
				pScrn->defaultVisual))
	    return FALSE;
    }
    else
    {
	if (!miSetVisualTypes(pScrn->depth,
			      miGetDefaultVisualMask(pScrn->depth),
			      pScrn->rgbBits, pScrn->defaultVisual))
	    return FALSE;
    }
#endif

    /*
     * Call the framebuffer layer's ScreenInit function, and fill in other
     * pScreen fields.
     */

    width = pScrn->virtualX;
    height = pScrn->virtualY;
    displayWidth = pScrn->displayWidth;

    if(pMQ->Rotate)
    {
	height = pScrn->virtualX;
	width = pScrn->virtualY;
    }

    if(pMQ->ShadowFB) {
 	pMQ->ShadowPitch = BitmapBytePad(pScrn->bitsPerPixel * width);
	pMQ->ShadowPtr = xalloc(pMQ->ShadowPitch * height);
	displayWidth = pMQ->ShadowPitch / (pScrn->bitsPerPixel >> 3);
        FBStart = pMQ->ShadowPtr;
    }
    else
    {
	pMQ->ShadowPtr = NULL;
        FBStart = pMQ->FbStart;
    }

    switch (pScrn->bitsPerPixel)
    {
	case 8:
	    ret = cfbScreenInit(pScreen, FBStart,
			width, height,
			pScrn->xDpi, pScrn->yDpi,
			displayWidth);
	    break;
	case 16:
	    ret = cfb16ScreenInit(pScreen, FBStart,
			width, height,
			pScrn->xDpi, pScrn->yDpi,
			displayWidth);
	    break;
	case 24:
	    if (pix24bpp == 24)
		ret = cfb24ScreenInit(pScreen, FBStart,
			width, height,
			pScrn->xDpi, pScrn->yDpi,
			displayWidth);
	    break;
	case 32:
	    ret = cfb32ScreenInit(pScreen, FBStart,
			width, height,
			pScrn->xDpi, pScrn->yDpi,
			displayWidth);
	    break;
	default:
	    xf86DrvMsg(scrnIndex, X_ERROR,
		   "Internal error: invalid bpp (%d) in MQScreenInit\n",
		   pScrn->bitsPerPixel);
	    ret = FALSE;
	    break;
    }
    if (!ret)
	return FALSE;

    if (pScrn->bitsPerPixel > 8)
    {
        /* Fixup RGB ordering */
        visual = pScreen->visuals + pScreen->numVisuals;
        while (--visual >= pScreen->visuals)
	{
	    if ((visual->class | DynamicClass) == DirectColor)
	    {
		visual->offsetRed = pScrn->offset.red;
		visual->offsetGreen = pScrn->offset.green;
		visual->offsetBlue = pScrn->offset.blue;
		visual->redMask = pScrn->mask.red;
		visual->greenMask = pScrn->mask.green;
		visual->blueMask = pScrn->mask.blue;
	    }
	}
    }

    xf86SetBlackWhitePixels(pScreen);


    //if(!pMQ->ShadowFB) /* hardware cursor needs to wrap this layer */
	//MQDGAInit(pScreen);

    if (!pMQ->NoAccel)
	MQAccelInit(pScreen);

    miInitializeBackingStore(pScreen);
    xf86SetBackingStore(pScreen);
    xf86SetSilkenMouse(pScreen);

    /* Initialize software cursor.  
	Must precede creation of the default colormap */
    miDCInitialize(pScreen, xf86GetPointerScreenFuncs());

    /* Initialize HW cursor layer. 
	Must follow software cursor initialization*/
    if (pMQ->HWCursor)
    { 
	if(!MQHWCursorInit(pScreen))
	    xf86DrvMsg(pScrn->scrnIndex, X_ERROR, 
		"Hardware cursor initialization failed\n");
    }

    /* Initialise default colourmap */
    if (!miCreateDefColormap(pScreen))
	return FALSE;

    /* Initialize colormap layer.  
	Must follow initialization of the default colormap */
    if(!xf86HandleColormaps(pScreen, 256, 8, 
	    //(pMQ->FBDev ? fbdevHWLoadPalette : MQdac->LoadPalette), NULL,
	    (pMQ->FBDev ? fbdevHWLoadPalette : MQLoadPalette), NULL,
	    CMAP_PALETTED_TRUECOLOR | CMAP_RELOAD_ON_MODE_SWITCH))	
	return FALSE;

    if(pMQ->ShadowFB)
    {
	RefreshAreaFuncPtr refreshArea = MQRefreshArea;

	if(pMQ->Rotate)
	{
	    if (!pMQ->PointerMoved)
	    {
		pMQ->PointerMoved = pScrn->PointerMoved;
		pScrn->PointerMoved = MQPointerMoved;
	    }

	   switch(pScrn->bitsPerPixel)
	   {
		case 8:	 refreshArea = MQRefreshArea8;	break;
		case 16: refreshArea = MQRefreshArea16;	break;
		case 24: refreshArea = MQRefreshArea24;	break;
		case 32: refreshArea = MQRefreshArea32;	break;
	   }
	}

	ShadowFBInit(pScreen, refreshArea);
    }

#ifdef DPMSExtension
    xf86DPMSInit(pScreen, MQDisplayPowerManagementSet, 0);
#endif
   
    pScrn->memPhysBase = pMQ->FbAddress;

    pScreen->SaveScreen = MQSaveScreen;

    /* Wrap the current CloseScreen function */
    pMQ->CloseScreen = pScreen->CloseScreen;
    pScreen->CloseScreen = MQCloseScreen;

    /* Report any unused options (only for the first generation) */
    if (serverGeneration == 1)
    {
	xf86ShowUnusedOptions(pScrn->scrnIndex, pScrn->options);
    }

    //ErrorF("MQScreenInit..done\n");
    /* Done */
    return TRUE;
}


/* Usually mandatory */
Bool
MQSwitchMode(int scrnIndex, DisplayModePtr mode, int flags)
{
    ErrorF("MQSwitchMode..\n");
    return MQModeInit(xf86Screens[scrnIndex], mode);
}


/*
 * This function is used to initialize the Start Address - the first
 * displayed location in the video memory.
 */
/* Usually mandatory */
void 
MQAdjustFrame(int scrnIndex, int x, int y, int flags)
{
    ScrnInfoPtr pScrn;
    int         Base, bytePP;
    //int         scrnHeight;
    //ULONG       stride;
    MQFBLayout *pLayout;
    MQPtr       pMQ;

    //ErrorF("MQAdjustFrame..%d,%d,%x\n",x,y,flags);
    pScrn = xf86Screens[scrnIndex];
    pMQ = MQPTR(pScrn);
    pLayout = &pMQ->CurrentLayout;
    bytePP = pLayout->bitsPerPixel / 8;
    //scrnHeight = pLayout->mode->VDisplay;
 
    if (pMQ->MQFlag & LARGE_DESKTOP)
    {
#if 0
	/* 4 possible layouts */
        stride = gcREAD(IW1_STRIDE) & 0xFFFF;
	switch (pMQ->flag & LCDCRT_POS_MASK)
	{
	    case HORI_CRT_LCD:
		start1 = (y * pLayout->displayWidth * bytePP)
                          + (x * bytePP);
		start2 = start1 + stride;
		break;

	    case HORI_LCD_CRT:
		start1 = (y * pLayout->displayWidth * bytePP)
                          + (x * bytePP);
		start2 = start1;
		start1 += stride;
		break;
				
	    case VERT_CRT_LCD:
		start1 = (y * pLayout->displayWidth * bytePP)
                          + (x * bytePP);
		start2 = start1 + stride * scrnHeight;
		break;
				
	    case VERT_LCD_CRT:
		start1 = startaddr + stride * scrnHeight;
		start2 = startaddr;
		break;
	}

	/* Program to the chip */
	REG32(IW1_STRIDE, stride);
	REG32(IW2_STRIDE, stride);

	REG32(IW1_START_ADDR, start1);
	REG32(IW2_START_ADDR, start2);
#endif
    }
    else
    {
	/* QView Same Image and Simultaneous LCD and/or CRT
	 *
	 * - set up 2 GCs in any cases ...
	 * - 2 addidtional registers write vs. code size
	 *
	 */

        //stride = gcREAD(IW1_STRIDE) & 0xFFFF;
        //stride = pScrn->displayWidth * bytePP;
	Base = (y * pLayout->displayWidth * bytePP) + (x * bytePP);
	gcREG(IW1_START_ADDR, Base);
	gcREG(IW2_START_ADDR, Base);
    }
}


/*
 * This is called when VT switching back to the X server.  Its job is
 * to reinitialise the video mode.
 *
 * We may wish to unmap video/MMIO memory too.
 */

/* Mandatory */
static Bool
MQEnterVT(int scrnIndex, int flags)
{
    ScrnInfoPtr pScrn = xf86Screens[scrnIndex];
     
    ErrorF("MQEnterVT..\n");
    if (!MQModeInit(pScrn, pScrn->currentMode))
	return FALSE;
    MQAdjustFrame(scrnIndex, pScrn->frameX0, pScrn->frameY0, 0);
    ErrorF("MQEnterVT..done\n");
    return TRUE;
}

static Bool
MQEnterVTFBDev(int scrnIndex, int flags)
{
    //ScrnInfoPtr pScrn = xf86Screens[scrnIndex];

    ErrorF("MQEnterVTFBDev..\n");
    fbdevHWEnterVT(scrnIndex,flags);
    ErrorF("MQEnterVTFBDev..done\n");
    return TRUE;
}

/*
 * This is called when VT switching away from the X server.  Its job is
 * to restore the previous (text) mode.
 *
 * We may wish to remap video/MMIO memory too.
 */

/* Mandatory */
static void
MQLeaveVT(int scrnIndex, int flags)
{
    ScrnInfoPtr pScrn = xf86Screens[scrnIndex];

    ErrorF("MQLeaveVT..\n");
    MQRestore(pScrn);
}


/*
 * This is called at the end of each server generation.  It restores the
 * original (text) mode.  It should also unmap the video memory, and free
 * any per-generation data allocated by the driver.  It should finish
 * by unwrapping and calling the saved CloseScreen function.
 */

/* Mandatory */
static Bool
MQCloseScreen(int scrnIndex, ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[scrnIndex];
    MQPtr pMQ = MQPTR(pScrn);

    ErrorF("MQCloseScreen..\n");
    if ( pScrn-> vtSema )
    {
        if (pMQ->FBDev)
        {
	    MQUnmapMemFBDev(pScrn);
        }
        else
        {
	    MQRestore(pScrn);
	    MQUnmapMem(pScrn);
        }
    }
 
#ifdef XF86DRI
/*
    if (pMQ->directRenderingEnabled)
    {
        MQDRICloseScreen(pScreen);
        pMQ->directRenderingEnabled=FALSE;
    }
*/
#endif

    if (pMQ->AccelInfoRec)
	XAADestroyInfoRec(pMQ->AccelInfoRec);
    if (pMQ->CursorInfoRec)
    	xf86DestroyCursorInfoRec(pMQ->CursorInfoRec);
    if (pMQ->ShadowPtr)
	xfree(pMQ->ShadowPtr);

    pScrn->vtSema = FALSE;

    pScreen->CloseScreen = pMQ->CloseScreen;
    return (*pScreen->CloseScreen)(scrnIndex, pScreen);
}


/* Free up any persistent data structures */

/* Optional */
static void
MQFreeScreen(int scrnIndex, int flags)
{
    /*
     * This only gets called when a screen is being deleted.  It does not
     * get called routinely at the end of a server generation.
     */
    MQFreeRec(xf86Screens[scrnIndex]);
}


/* Checks if a mode is suitable for the selected chipset. */

/* Optional */
static int
MQValidMode(int scrnIndex, DisplayModePtr mode, Bool verbose, int flags)
{
    ErrorF("MQValidMode..\n");

    return(MODE_OK);
}


/* Do screen blanking */

/* Mandatory */
static Bool
MQSaveScreen(ScreenPtr pScreen, int mode)
{
    return TRUE;
}


/*
 * MQDisplayPowerManagementSet --
 *
 * Sets VESA Display Power Management Signaling (DPMS) Mode.
 */
#ifdef DPMSExtension

#define DPMSModeOn	0
#define DPMSModeStandby	1
#define DPMSModeSuspend	2
#define DPMSModeOff	3


static void
MQDisplayPowerManagementSet(ScrnInfoPtr pScrn, int PowerManagementMode,
			     int flags)
{
    MQPtr pMQ = MQPTR(pScrn);
    ULONG regdata;

    ErrorF("MQDisplayPowerManagementSet: %d\n", PowerManagementMode);
    switch( PowerManagementMode )
    {
 	case DPMSModeOn:
	    /* Screen: On; HSync: On, VSync: On */
            regdata = pciREAD(PCI_PM_CNTL_STATUS);
	    if ((regdata & 3) > 0)
                pciREG(PCI_PM_CNTL_STATUS, ENTER_D0);
	    break;
	case DPMSModeStandby:
	    /* Screen: Off; HSync: Off, VSync: On */
            pciREG(PCI_PM_CNTL_STATUS, ENTER_D1);
	    break;
	case DPMSModeSuspend:
	    /* Screen: Off; HSync: On, VSync: Off */
            pciREG(PCI_PM_CNTL_STATUS, ENTER_D2);
	    break;
	case DPMSModeOff:
	    /* Screen: Off; HSync: Off, VSync: Off */
            pciREG(PCI_PM_CNTL_STATUS, ENTER_D3);
	    break;
   }
   sleep( 5 );
}
#endif
